package org.kantega.jexmec.manager;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.*;
import org.kantega.jexmec.events.PluginClassLoaderEvent;
import org.kantega.jexmec.events.PluginLoadingExceptionEvent;
import org.kantega.jexmec.events.PluginManagerEvent;
import org.kantega.jexmec.events.PluginRegistrationEvent;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.net.URL;
import java.util.*;

/**
 * Default implementation of PluginManager. Tracks the ClassLoaders registered by {@link PluginClassLoaderProvider}s and uses the provided
 * {@link org.kantega.jexmec.PluginLoader} implementations with {@link org.kantega.jexmec.ServiceLocator}s to load plugins from the ClassLoaders.
 */
public final class DefaultPluginManager<P> implements PluginManager<P> {

    private final Class<P> pluginClass;

    private final List<PluginLoader<P>> pluginLoaders = new ArrayList<PluginLoader<P>>();

    private final List<PluginClassLoaderProvider> pluginClassLoaderProviders = new ArrayList<PluginClassLoaderProvider>();

    private final List<ServiceLocator> serviceLocators = new ArrayList<ServiceLocator>();

    private final List<PluginManagerListener<P>> pluginManagerListeners = Collections.synchronizedList(new ArrayList<PluginManagerListener<P>>());

    private final ServiceLocator serviceLocator = new CompoundServiceLocator(serviceLocators);

    private final PluginRegistry registry = new PluginRegistry(serviceLocator);

    private volatile State state = State.UNSTARTED;

    private final List<PluginClassLoaderProvider> defaultPluginClassLoaderProviders = new ArrayList<PluginClassLoaderProvider>();

    private boolean rethrowPluginLoaderExceptions = true;

    /**
     * Creates a DefaultPluginManager managing the given type of plugins
     *
     * @param pluginClass the class of the plugin to manage.
     * @throws IllegalArgumentException if the plugin class is not an interface
     */
    public DefaultPluginManager(Class<P> pluginClass) {
        if(!pluginClass.isInterface()) {
            throw new IllegalArgumentException("Plugin class needs to be an interface: " + pluginClass);
        }
        this.pluginClass = pluginClass;
        defaultPluginClassLoaderProviders.add(new StaticPluginClassLoaderProvider(getClass().getClassLoader()));
    }

    /**
     * Adds a {@link org.kantega.jexmec.PluginClassLoaderProvider} to the plugin manager.
     *
     * @param pluginClassLoaderProvider the provider to add
     * @return the plugin manager
     */
    public DefaultPluginManager<P> addPluginClassLoaderProvider(PluginClassLoaderProvider pluginClassLoaderProvider) {
        state.assertEquals(State.UNSTARTED, "adding plugin class loader provider " + pluginClassLoaderProvider);
        defaultPluginClassLoaderProviders.clear();
        pluginClassLoaderProviders.add(pluginClassLoaderProvider);
        return this;
    }

    /**
     * Adds a plugin classloader to the plugin manager. This method calls
     * {@link #addPluginClassLoaderProvider(org.kantega.jexmec.PluginClassLoaderProvider)} with
     * {@link PluginClassLoaderProvider} providing the classloader
     *
     * @param pluginClassloader the classloader to add a provider for
     * @return the plugin manager
     */
    public DefaultPluginManager<P> addPluginClassLoader(ClassLoader pluginClassloader) {
        return addPluginClassLoaderProvider(new StaticPluginClassLoaderProvider(pluginClassloader));
    }

    /**
     * Adds a {@link org.kantega.jexmec.PluginLoader} to the plugin manager.
     *
     * @param pluginLoader the loader to add
     * @return the plugin manager
     */
    public DefaultPluginManager<P> addPluginLoader(PluginLoader<P> pluginLoader) {
        state.assertEquals(State.UNSTARTED, "adding plugin loader " + pluginLoader);
        pluginLoaders.add(pluginLoader);
        return this;
    }

    /**
     * Adds a {@link org.kantega.jexmec.ServiceLocator} to the plugin manager.
     *
     * @param serviceLocator the locator to add
     * @return the plugin manager
     */
    public DefaultPluginManager<P> addServiceLocator(ServiceLocator serviceLocator) {
        state.assertEquals(State.UNSTARTED, "adding service locator " + serviceLocator);
        serviceLocators.add(serviceLocator);
        return this;
    }

    /**
     * Adds a single service mapping. Shortcut for calling addServiceLocator(new StaticServiceLocator(ServiceKey, Object))
     *
     * @param serviceKey the key to add a service for
     * @param service    the service implementation
     * @return the plugin manager.
     */
    public <T> DefaultPluginManager<P> addService(ServiceKey<T> serviceKey, T service) {
        return addServiceLocator(new StaticServiceLocator(serviceKey, service));
    }

    /**
     * Creates an instance of a services interface. The instance is implemented as
     * a dynamic proxy that will look up services the service locators added to this plugin manager
     *
     * @param servicesClass the service class to implement
     * @param <S>           the service class type
     * @return an instance of the service interface
     */
    public <S> S createServicesInstance(final Class<S> servicesClass) {
        return servicesClass.cast(Proxy.newProxyInstance(DefaultPluginManager.class.getClassLoader(),
                new Class[]{servicesClass},
                new InvocationHandler() {
                    public Object invoke(Object o, Method method, Object[] objects) throws Throwable {
                        ServiceKey<?> key = objects == null ?
                                ServiceKey.by(method.getReturnType()) : ServiceKey.by(method.getReturnType(), ((Enum) objects[0]).name());

                        return serviceLocator.lookupService(key);
                    }
                }));
    }

    /**
     * Creates a classloader hiding resource under <code>META-INF/services/MyPluginClass</code>
     * and <code>META-INF/services/com.mycompany.MyPluginClass</code.
     * @param parent the parent class loader
     * @return the class loader hiding resource in the parent class loader
     */
    ClassLoader createPluginResourceHidingClassLoader(ClassLoader parent) {
        return new ResourceHidingClassLoader(parent, pluginClass);
    }

    /**
     * <p>Creates an {@link Iterable} of plugins that is a dynamic proxy to {@link DefaultPluginManager#getPlugins()}.
     * Adding the plugin collection as a service gives plugins access to other plugins, but without the Jexmec
     * compile dependency which adding DefaultPluginManager would introduce.</p>
     * <p/>
     * <p>
     * To use this feature, add an interface
     * <code>public interface Plugins extends Iterable&lt;MyPlugin&gt; {}</code> to your plugin API. The interface should not
     * add any methods, it is just created for type safety. Then create an
     * instance of Plugins: <code>Plugins plugins = pluginManager.createPluginIterable(Plugins.class);</code>.
     * Finally, add the service to DefaultPluginManager: <code>pluginManager.addService(ServiceKey.by(PluginList.class, pluginList));</code>
     * </p>
     *
     * @param pluginCollectionClass As interface extending Collection&lt;P;&gt; where P is your plugin class
     * @param <P>                   the plugin type
     * @param <T>                   the collection type
     * @return the collection instance
     */
    @SuppressWarnings("unchecked")
    public <P, T extends Iterable<P>> T createPluginIterable(Class<T> pluginCollectionClass) {

        return (T) Proxy.newProxyInstance(DefaultPluginManager.class.getClassLoader(), new Class[]{pluginCollectionClass}, new InvocationHandler() {
            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                return method.invoke(method.getName().equals("iterator") ? getPlugins() : this, args);
            }
        });
    }


    /**
     * Adds any default plugin loaders or plugin class loader providers.
     */
    private void addDefaults() {
        pluginClassLoaderProviders.addAll(defaultPluginClassLoaderProviders);
    }

    /**
     * Starts the plugin manager.
     *
     * @return the plugin manager
     */
    public synchronized DefaultPluginManager<P> start() {
        state.assertEquals(State.UNSTARTED, "starting plugin manager");

        for (PluginManagerListener<P> listener : pluginManagerListeners) {
            listener.beforePluginManagerStarted(new DefaultPluginManagerEvent<P>(this));
        }
        state = State.STARTING;

        addDefaults();

        startPluginLoaders();

        startPluginClassLoaderProviders();

        state = State.STARTED;

        for (PluginManagerListener<P> listener : pluginManagerListeners) {
            listener.afterPluginManagerStarted(new DefaultPluginManagerEvent<P>(this));
        }

        return this;
    }

    /**
     * Stops the plugin manager.
     *
     * @return the plugin manager.
     */
    public synchronized DefaultPluginManager<P> stop() {
        state.assertEquals(State.STARTED, "stopping plugin manager");

        for (PluginManagerListener<P> listener : pluginManagerListeners) {
            listener.beforePluginManagerStopped(new DefaultPluginManagerEvent<P>(this));
        }

        state = State.STOPPING;

        stopPluginClassLoaderProviders();

        stopPluginLoaders();

        state = State.STOPPED;

        for (PluginManagerListener<P> listener : pluginManagerListeners) {
            listener.afterPluginManagerStopped(new DefaultPluginManagerEvent<P>(this));
        }
        return this;
    }


    /**
     * Start all plugin classloader providers
     */
    private void startPluginClassLoaderProviders() {
        ClassLoader parentClassLoader = new ResourceHidingClassLoader(getClass().getClassLoader(), pluginClass);
        for (PluginClassLoaderProvider provider : pluginClassLoaderProviders) {
            provider.start(new ProviderAwarePluginClassLoaderRegistry(provider, registry), parentClassLoader);
        }
    }

    /**
     * Starts all plugin loaders
     */
    private void startPluginLoaders() {
        for (PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.start();
        }
    }

    /**
     * Stops all plugin loaders.
     */
    private void stopPluginLoaders() {
        for (PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.stop();
        }
    }

    /**
     * Stops all plugin classloader providers.
     */
    private void stopPluginClassLoaderProviders() {
        for (PluginClassLoaderProvider provider : pluginClassLoaderProviders) {
            provider.stop();
        }
    }

    /**
     * Return a list of the plugins currently installed in this PluginManager
     * @return  the list of plugins
     */

    public List<P> getPlugins() {
        state.assertEquals(State.STARTED, "getting plugins. You need to call start() *before* calling getPlugins()");
        return Collections.unmodifiableList(registry.getPlugins());
    }


    /**
     * Register a PluginManagerListener for event notification.
     * @param listener the listener to add
     */
    public void addPluginManagerListener(PluginManagerListener<P> listener) {
        pluginManagerListeners.add(listener);
    }

    /**
     * Unregister a PluginManagerListener for event notification.
     * @param listener the listener to remove
     */
    public void removePluginManagerListener(PluginManagerListener<P> listener) {
        pluginManagerListeners.remove(listener);
    }

    /**
     * Return the ClassLoader that initiated the loading of the given plugin. Note that this
     * need not be the same as <code>plugin.getClassLoader()</code>
     * @param plugin the plugin for which to return the loading plugin
     * @return the initiating ClassLoader
     */
    public ClassLoader getClassLoader(P plugin) {
        return registry.getClassLoader(plugin);
    }

    /**
     * If set to false, exceptions thrown when loading a plugin from a PluginLoader will not be rethrown.
     * Useful when one failing plugin should not stop other plugins from loading.
     * Please not that when this is set to false AND no PluginManagerListener reports exceptions
     * by implementing {@link PluginManagerListener#pluginLoadingFailedWithException(org.kantega.jexmec.events.PluginLoadingExceptionEvent)},
     * exceptions during plugin loading will be effectively swallowed.
     * The default for this value is true.
     * @param rethrowPluginLoaderExceptions true if exceptions should be rethrown, false otherwise.
     */
    public void setRethrowPluginLoaderExceptions(boolean rethrowPluginLoaderExceptions) {
        this.rethrowPluginLoaderExceptions = rethrowPluginLoaderExceptions;
    }

    /**
     * A registry keeping track of plugins.
     */
    private class PluginRegistry implements PluginClassLoaderProvider.Registry {

        private final ServiceLocator serviceLocator;

        private final Map<ClassLoader, Map<PluginLoader<P>, List<P>>> plugins = new LinkedHashMap<ClassLoader, Map<PluginLoader<P>, List<P>>>();

        private final Map<P, ClassLoader> classLoaderMap = new HashMap<P, ClassLoader>();


        public PluginRegistry(ServiceLocator serviceLocator) {
            this.serviceLocator = serviceLocator;
        }

        public <T extends ClassLoader> void replace(final Iterable<T> removeClassLoaders, Iterable<T> addClassLoaders) {

            PluginClassLoaderProvider provider = ProviderAwarePluginClassLoaderRegistry.getCurrentProvider();

            {
                // Validate that all class loaders to remove are actually registered
                synchronized (this) {
                    for (ClassLoader loader : removeClassLoaders) {
                        if(!plugins.containsKey(loader)) {
                            throw new IllegalArgumentException("Illegal request to remove a class loader that isn't currently registered: " + loader);
                        }
                    }
                }
                for (ClassLoader loader : removeClassLoaders) {
                    for (PluginManagerListener<P> listener : pluginManagerListeners) {
                        listener.beforeClassLoaderRemoved(new DefaultPluginClassLoaderEvent<P>(DefaultPluginManager.this, provider, loader));
                    }

                    Map<PluginLoader<P>, List<P>> pluginsRemoved;
                    synchronized (this) {
                        pluginsRemoved = plugins.get(loader);
                    }

                    for(PluginLoader<P> pluginLoader : pluginsRemoved.keySet()) {
                        for (PluginManagerListener<P> listener : pluginManagerListeners) {
                            listener.beforePassivation(new DefaultPluginRegistrationEvent<P>(DefaultPluginManager.this, provider, loader, pluginLoader, pluginsRemoved.get(pluginLoader)));
                        }
                    }

                    synchronized (this) {
                        for(PluginLoader<P> pluginLoader : pluginsRemoved.keySet()) {
                            for (P plugin : pluginsRemoved.get(pluginLoader)) {
                                classLoaderMap.remove(plugin);
                            }
                        }
                        plugins.remove(loader);

                    }

                    for(PluginLoader<P> pluginLoader : pluginsRemoved.keySet()) {
                        for (PluginManagerListener<P> listener : pluginManagerListeners) {
                            listener.afterPassivation(new DefaultPluginRegistrationEvent<P>(DefaultPluginManager.this, provider, loader, pluginLoader, pluginsRemoved.get(pluginLoader)));
                        }
                    }


                    unloadPluginsFromClassLoader(loader);

                    for (PluginManagerListener<P> listener : pluginManagerListeners) {
                        listener.afterClassLoaderRemoved(new DefaultPluginClassLoaderEvent<P>(DefaultPluginManager.this, provider, loader));
                    }
                }

            }

            {

                for (ClassLoader classLoader : addClassLoaders) {
                    for (PluginManagerListener<P> listener : pluginManagerListeners) {
                        listener.beforeClassLoaderAdded(new DefaultPluginClassLoaderEvent<P>(DefaultPluginManager.this, provider, classLoader));
                    }



                    Map<PluginLoader<P>, List<P>> loadedPlugins = new LinkedHashMap<PluginLoader<P>,List<P>>();

                    for (PluginLoader<P> pluginLoader : pluginLoaders) {
                        try {
                            loadedPlugins.put(pluginLoader, pluginLoader.loadPlugins(pluginClass, classLoader, serviceLocator));
                        } catch(Throwable e) {
                            fireExceptionLoadingPlugins(provider, classLoader, pluginLoader, e);
                            if(rethrowPluginLoaderExceptions) {
                                {
                                    throw new RuntimeException("Plugin loader " + pluginLoader + " threw " + e.getClass().getSimpleName() +" loading plugins from class loader " + classLoader, e);
                                }
                            }
                        }
                    }

                    for(PluginLoader<P> pluginLoader : loadedPlugins.keySet()) {
                        if(loadedPlugins.get(pluginLoader).size() > 0) {
                            for (PluginManagerListener<P> listener : pluginManagerListeners) {
                                listener.beforeActivation(new DefaultPluginRegistrationEvent<P>(DefaultPluginManager.this, provider, classLoader, pluginLoader, loadedPlugins.get(pluginLoader)));
                            }
                        }
                    }

                    synchronized (this) {
                        List<P> plugins  = new ArrayList<P>();

                        for(PluginLoader<P> pluginLoader : loadedPlugins.keySet()) {
                            plugins.addAll(loadedPlugins.get(pluginLoader));
                        }
                        for (P plugin : plugins) {
                            classLoaderMap.put(plugin, classLoader);
                        }

                        this.plugins.put(classLoader, loadedPlugins);
                    }

                    for(PluginLoader<P> pluginLoader : loadedPlugins.keySet()) {
                        if(loadedPlugins.get(pluginLoader).size() > 0) {
                            for (PluginManagerListener<P> listener : pluginManagerListeners) {
                                listener.afterActivation(new DefaultPluginRegistrationEvent<P>(DefaultPluginManager.this, provider, classLoader, pluginLoader, loadedPlugins.get(pluginLoader)));
                            }
                        }
                    }

                    for (PluginManagerListener<P> listener : pluginManagerListeners) {
                        listener.afterClassLoaderAdded(new DefaultPluginClassLoaderEvent<P>(DefaultPluginManager.this, provider, classLoader));
                    }
                }
            }
            firePluginsUpdated();


        }

        public <T extends ClassLoader> void add(Iterable<T> classLoaders) {
            replace(Collections.<T>emptySet(), classLoaders);
        }

        public <T extends ClassLoader> void remove(Iterable<T> loaders) {
            replace(loaders, Collections.<T>emptySet());
        }


        /**
         * Notify listeners that an exception was thrown during plugin loading.
         * @param provider the plugin class loader provider which requested classloader to be added
         * @param classLoader the class loader which the plugin loader was loading
         * @param pluginLoader the plugin loader which loading failed
         * @param e the Throwable exception or error
         */
        private void fireExceptionLoadingPlugins(PluginClassLoaderProvider provider, ClassLoader classLoader, PluginLoader<P> pluginLoader, Throwable e) {
            for(PluginManagerListener<P> listener : pluginManagerListeners) {
                listener.pluginLoadingFailedWithException(new DefaultPluginLoadingExceptionEvent<P>(DefaultPluginManager.this, provider, classLoader,  pluginLoader, e));
            }
        }

        /**
         * Tell all plugin loaders to unload plugins loaded from the given class loader.
         *
         * @param loader the ClassLoader for which all plugins should be unloaded.
         */
        private void unloadPluginsFromClassLoader(ClassLoader loader) {
            for (PluginLoader<P> pluginLoader : pluginLoaders) {
                pluginLoader.unloadPlugins(loader);
            }
        }

        /**
         * Return currently installed plugins
         *
         * @return a list of installed plugins
         */
        public List<P> getPlugins() {
            List<P> plugins = new ArrayList<P>();

            synchronized (this) {
                for(Map<PluginLoader<P>, List<P>> map : this.plugins.values()) {
                    for(PluginLoader<P> pluginLoader : map.keySet()) {
                        plugins.addAll(map.get(pluginLoader));
                    }
                }
            }
            return plugins;
        }

        /**
         * Returns the classloader that caused the loading of the given plugin.
         *
         * @param plugin the plugin
         * @return the classloader which the plugin was loaded from
         */
        public ClassLoader getClassLoader(P plugin) {
            return classLoaderMap.get(plugin);
        }

        private void firePluginsUpdated() {
            for (PluginManagerListener<P> listener : pluginManagerListeners) {
                listener.pluginsUpdated(registry.getPlugins());
            }
        }
    }

    /**
     * A service locator that wraps a collecton of service locators. Services are looked up in
     * the collection order. The first non null service is returned. If no services are found, service
     * lookup returns null.
     */
    private class CompoundServiceLocator implements ServiceLocator {
        private final Collection<ServiceLocator> serviceLocators;

        public CompoundServiceLocator(Collection<ServiceLocator> serviceLocators) {
            this.serviceLocators = serviceLocators;
        }

        /**
         * Gets the the set of service keys supported by the collection
         * of service locators.
         *
         * @return the combined set of service keys
         */
        public Set<ServiceKey> getServiceKeys() {
            Set<ServiceKey> serviceKeys = new HashSet<ServiceKey>();
            for (ServiceLocator serviceLocator : serviceLocators) {
                serviceKeys.addAll(serviceLocator.getServiceKeys());
            }
            return serviceKeys;

        }

        /**
         * Returns the first non-null service implementation matching the given key. If no service is
         * found, null is returned.
         *
         * @param serviceKey the service key class to find an implementation for
         * @param <T>        the type of the service
         * @return the service instance or null if the service was not found.
         */
        public <T> T lookupService(ServiceKey<T> serviceKey) {
            ServiceLocator first = null;
            for (Iterator<ServiceLocator> i = serviceLocators.iterator(); first == null && i.hasNext();) {
                ServiceLocator candidate = i.next();
                first = candidate.getServiceKeys().contains(serviceKey) ? candidate : null;

            }
            return first != null ? first.lookupService(serviceKey) : null;
        }
    }

    private static class StaticServiceLocator implements ServiceLocator {

        private final Map<ServiceKey, Object> services;


        public StaticServiceLocator(ServiceKey serviceKey, Object service) {
            this.services = Collections.singletonMap(serviceKey, service);
        }

        public Set<ServiceKey> getServiceKeys() {
            return services.keySet();
        }

        public <T> T lookupService(ServiceKey<T> serviceKey) {
            return serviceKey.getServiceType().cast(services.get(serviceKey));
        }
    }

    private static class StaticPluginClassLoaderProvider implements PluginClassLoaderProvider {
        private final List<ClassLoader> classLoaders = new ArrayList<ClassLoader>();
        private Registry registry;

        public StaticPluginClassLoaderProvider(ClassLoader... classLoaders) {
            this.classLoaders.addAll(Arrays.asList(classLoaders));
        }

        public void start(Registry registry, ClassLoader parentClassLoader) {
            this.registry = registry;
            registry.add(classLoaders);
        }

        public void stop() {
            registry.remove(classLoaders);
        }
    }

    class ResourceHidingClassLoader extends ClassLoader {

        private final String[] localResourcePrefixes;

        /**
         * Creates a ResourceHidingClassLoader hiding resources in <code>META-INF/services/PluginName/</code> and
         * <code>META-INF/services/com.example.PluginName/</code>.
         *
         * @param parent      the parent class loader
         * @param pluginClass the plugin class to hide resources for.
         */
        public ResourceHidingClassLoader(ClassLoader parent, Class pluginClass) {
            super(parent);
            localResourcePrefixes = new String[]{"META-INF/services/" + pluginClass.getSimpleName() + "/",
                    "META-INF/services/" + pluginClass.getName() + "/"};
        }


        @Override
        public InputStream getResourceAsStream(String name) {
            final URL resource = getResource(name);
            try {
                return resource == null ? null : resource.openStream();
            } catch (IOException e) {
                return null;
            }
        }

        @Override
        public URL getResource(String name) {
            if (isLocalResource(name)) {
                return super.findResource(name);
            } else {
                return super.getResource(name);
            }
        }

        @Override
        public Enumeration<URL> getResources(String name) throws IOException {
            if (isLocalResource(name)) {
                return super.findResources(name);
            } else {
                return super.getResources(name);
            }
        }

        protected boolean isLocalResource(String name) {
            for (String localResourcePrefix : localResourcePrefixes) {
                if (name.startsWith(localResourcePrefix)) {
                    return true;
                }
            }
            return false;
        }
    }

    private static class ProviderAwarePluginClassLoaderRegistry implements PluginClassLoaderProvider.Registry  {
        private static final ThreadLocal<PluginClassLoaderProvider> providerThreadLocal = new ThreadLocal<PluginClassLoaderProvider>();
        private final PluginClassLoaderProvider wrappedProvider;
        private final PluginClassLoaderProvider.Registry wrappedRegistry;

        public ProviderAwarePluginClassLoaderRegistry(PluginClassLoaderProvider wrappedProvider, PluginClassLoaderProvider.Registry wrappedRegistry) {
            this.wrappedProvider = wrappedProvider;
            this.wrappedRegistry = wrappedRegistry;
        }

        public <T extends ClassLoader> void add(Iterable<T> classLoaders) {
            providerThreadLocal.set(wrappedProvider);
            try {
                wrappedRegistry.add(classLoaders);
            } finally {
                providerThreadLocal.remove();
            }
        }

        public <T extends ClassLoader> void remove(Iterable<T> classLoaders) {
            providerThreadLocal.set(wrappedProvider);
            try {
                wrappedRegistry.remove(classLoaders);
            } finally {
                providerThreadLocal.remove();
            }
        }

        public <T extends ClassLoader> void replace(Iterable<T> removeClassLoaders, Iterable<T> addClassLoaders) {
            try {
                providerThreadLocal.set(wrappedProvider);
                wrappedRegistry.replace(removeClassLoaders, addClassLoaders);
            } finally {
                providerThreadLocal.remove();
            }
        }

        public static PluginClassLoaderProvider getCurrentProvider() {
            return providerThreadLocal.get();
        }
    }

    private enum State {
        UNSTARTED, STARTING, STARTED, STOPPING, STOPPED;

        /**
         * Throws an IllegalStateException if the <code>actual<code> state is different from the <code>expected</code> state.
         *
         * @param expected the expected state
         * @param when     description of when this state assertion failed
         */
        public void assertEquals(State expected, String when) {
            if (this != expected) {
                throw new IllegalStateException("Illegal state " + this + ". Expected " + expected + " when " + when);
            }
        }
    }

    static class DefaultPluginManagerEvent<P> implements PluginManagerEvent<P> {
        private final DefaultPluginManager<P> pluginManger;

        public DefaultPluginManagerEvent(DefaultPluginManager<P> pluginManger) {
            this.pluginManger = pluginManger;
        }

        public DefaultPluginManager<P> getPluginManager() {
            return pluginManger;
        }
    }

    static class DefaultPluginClassLoaderEvent<P> extends DefaultPluginManagerEvent<P> implements PluginClassLoaderEvent<P> {
        private final PluginClassLoaderProvider provider;
        private final ClassLoader classLoader;

        public DefaultPluginClassLoaderEvent(DefaultPluginManager<P> pluginManger, PluginClassLoaderProvider provider, ClassLoader classLoader) {
            super(pluginManger);
            this.provider = provider;
            this.classLoader = classLoader;
        }


        public PluginClassLoaderProvider getProvider() {
            return provider;
        }

        public ClassLoader getClassLoader() {
            return classLoader;
        }
    }

    static class DefaultPluginRegistrationEvent<P> extends DefaultPluginClassLoaderEvent<P> implements PluginRegistrationEvent<P> {

        private final List<P> plugins;

        private final PluginLoader<P> pluginLoader;

        DefaultPluginRegistrationEvent(DefaultPluginManager<P> pluginManger, PluginClassLoaderProvider provider, ClassLoader classLoader, PluginLoader<P> pluginLoader, List<P> plugins) {
            super(pluginManger, provider, classLoader);
            this.plugins = plugins;
            this.pluginLoader = pluginLoader;
        }

        public List<P> getPlugins() {
            return this.plugins;
        }

        public PluginLoader<P> getPluginLoader() {
            return pluginLoader;
        }
    }

    static class DefaultPluginLoadingExceptionEvent<P> extends DefaultPluginClassLoaderEvent<P> implements PluginLoadingExceptionEvent<P> {
        private final PluginLoader<P> pluginLoader;
        private final Throwable throwable;

        DefaultPluginLoadingExceptionEvent(DefaultPluginManager<P> pluginManger, PluginClassLoaderProvider provider, ClassLoader classLoader, PluginLoader<P> pluginLoader, Throwable throwable) {
            super(pluginManger, provider, classLoader);
            this.pluginLoader = pluginLoader;
            this.throwable = throwable;
        }

        public PluginLoader<P> getPluginLoader() {
            return pluginLoader;
        }

        public Throwable getThrowable() {
            return throwable;
        }
    }


}
